﻿Imports System.Threading
Imports System.Runtime.InteropServices
Imports System.Printing
Imports Microsoft.Win32.SafeHandles

Public Class PrintJobChangeEventArgs
    Inherits EventArgs

    Public Sub New(intJobID As Integer, strJobName As String, jStatus As JOBSTATUS, objJobInfo As PrintSystemJobInfo)
        MyBase.New()
        _JobID = intJobID
        _JobName = strJobName
        _JobStatus = jStatus
        _JobInfo = objJobInfo
    End Sub

    Public ReadOnly Property JobID() As Integer

    Public ReadOnly Property JobName() As String

    Public ReadOnly Property JobStatus() As JOBSTATUS

    Public ReadOnly Property JobInfo() As PrintSystemJobInfo

End Class

Public Delegate Sub PrintJobStatusChanged(Sender As Object, e As PrintJobChangeEventArgs)

Public Class PrintQueueMonitor

#Region "   DLL Import Functions "

    <DllImport("winspool.drv", EntryPoint:="OpenPrinterA", SetLastError:=True, CharSet:=CharSet.Ansi, ExactSpelling:=True, CallingConvention:=CallingConvention.StdCall)>
    Public Shared Function OpenPrinter(pPrinterName As [String], ByRef phPrinter As IntPtr, pDefault As Int32) As Boolean
    End Function

    <DllImport("winspool.drv", EntryPoint:="ClosePrinter", SetLastError:=True, ExactSpelling:=True, CallingConvention:=CallingConvention.StdCall)>
    Public Shared Function ClosePrinter(hPrinter As Int32) As Boolean
    End Function

    <DllImport("winspool.drv", EntryPoint:="FindFirstPrinterChangeNotification", SetLastError:=True, CharSet:=CharSet.Ansi, ExactSpelling:=True, CallingConvention:=CallingConvention.StdCall)>
    Public Shared Function FindFirstPrinterChangeNotification(<InAttribute> hPrinter As IntPtr, <InAttribute> fwFlags As Int32, <InAttribute> fwOptions As Int32, <InAttribute, MarshalAs(UnmanagedType.LPStruct)> pPrinterNotifyOptions As PRINTER_NOTIFY_OPTIONS) As IntPtr
    End Function

    <DllImport("winspool.drv", EntryPoint:="FindNextPrinterChangeNotification", SetLastError:=True, CharSet:=CharSet.Ansi, ExactSpelling:=False, CallingConvention:=CallingConvention.StdCall)>
    Public Shared Function FindNextPrinterChangeNotification(<InAttribute> hChangeObject As IntPtr, <OutAttribute> ByRef pdwChange As Int32, <InAttribute, MarshalAs(UnmanagedType.LPStruct)> pPrinterNotifyOptions As PRINTER_NOTIFY_OPTIONS, <OutAttribute> ByRef lppPrinterNotifyInfo As IntPtr) As Boolean
    End Function

#End Region

#Region "   Constants"

    Const PRINTER_NOTIFY_OPTIONS_REFRESH As Integer = 1

#End Region

#Region "   Events "

    Public Event OnJobStatusChange As PrintJobStatusChanged

#End Region

#Region "private variables"
    Private _printerHandle As IntPtr = IntPtr.Zero
    Private _spoolerName As String = ""
    Private _mrEvent As New ManualResetEvent(False)
    Private _waitHandle As RegisteredWaitHandle = Nothing
    Private _changeHandle As IntPtr = IntPtr.Zero
    Private _notifyOptions As New PRINTER_NOTIFY_OPTIONS()
    Private objJobDict As New Dictionary(Of Integer, String)()
    Private _spooler As PrintQueue = Nothing
#End Region

#Region "   constructor "

    Public Sub New(strSpoolName As String)
        ' Let us open the printer and get the printer handle.
        _spoolerName = strSpoolName
        'Start Monitoring
        Start()
    End Sub

#End Region

#Region "   destructor"

    Protected Overrides Sub Finalize()
        Try
            [Stop]()
        Finally
            MyBase.Finalize()
        End Try
    End Sub

#End Region

#Region "   StartMonitoring "
    Public Sub Start()
        OpenPrinter(_spoolerName, _printerHandle, 0)
        If _printerHandle <> IntPtr.Zero Then
            'We got a valid Printer handle.  Let us register for change notification....
            _changeHandle = FindFirstPrinterChangeNotification(_printerHandle, CType(PRINTER_CHANGES.PRINTER_CHANGE_JOB, Integer), 0, _notifyOptions)
            ' We have successfully registered for change notification.  Let us capture the handle...
            _mrEvent.SafeWaitHandle = New SafeWaitHandle(_changeHandle, False)
            '_mrEvent.Handle = _changeHandle
            'Now, let us wait for change notification from the printer queue....
            _waitHandle = ThreadPool.RegisterWaitForSingleObject(_mrEvent, New WaitOrTimerCallback(AddressOf PrinterNotifyWaitCallback), _mrEvent, -1, False)
        End If

        _spooler = New PrintQueue(New PrintServer(), _spoolerName)
        For Each psi As PrintSystemJobInfo In _spooler.GetPrintJobInfoCollection()
            objJobDict(psi.JobIdentifier) = psi.Name
        Next
    End Sub
#End Region

#Region "   StopMonitoring "
    Public Sub [Stop]()
        If _printerHandle <> IntPtr.Zero Then
            ClosePrinter(CType(_printerHandle, Integer))
            _printerHandle = IntPtr.Zero
        End If
    End Sub
#End Region

#Region "Callback Function"
    Public Sub PrinterNotifyWaitCallback(state As [Object], timedOut As Boolean)
        If _printerHandle = IntPtr.Zero Then
            Return
        End If
        '#Region "read notification details"
        _notifyOptions.Count = 1
        Dim pdwChange As Integer = 0
        Dim pNotifyInfo As IntPtr = IntPtr.Zero
        Dim bResult As Boolean = FindNextPrinterChangeNotification(_changeHandle, pdwChange, _notifyOptions, pNotifyInfo)
        'If the Printer Change Notification Call did not give data, exit code
        If (bResult = False) OrElse ((CType(pNotifyInfo, Integer)) = 0) Then
            Return
        End If

        'If the Change Notification was not relgated to job, exit code
        Dim bJobRelatedChange As Boolean = ((pdwChange And PRINTER_CHANGES.PRINTER_CHANGE_ADD_JOB) = PRINTER_CHANGES.PRINTER_CHANGE_ADD_JOB) OrElse ((pdwChange And PRINTER_CHANGES.PRINTER_CHANGE_SET_JOB) = PRINTER_CHANGES.PRINTER_CHANGE_SET_JOB) OrElse ((pdwChange And PRINTER_CHANGES.PRINTER_CHANGE_DELETE_JOB) = PRINTER_CHANGES.PRINTER_CHANGE_DELETE_JOB) OrElse ((pdwChange And PRINTER_CHANGES.PRINTER_CHANGE_WRITE_JOB) = PRINTER_CHANGES.PRINTER_CHANGE_WRITE_JOB)
        If Not bJobRelatedChange Then
            Return
        End If
#End Region

#Region "populate Notification Information"
        'Now, let us initialize and populate the Notify Info data
        Dim info As PRINTER_NOTIFY_INFO = CType(Marshal.PtrToStructure(pNotifyInfo, GetType(PRINTER_NOTIFY_INFO)), PRINTER_NOTIFY_INFO)
        Dim pData As Integer = CType(pNotifyInfo, Integer) + Marshal.SizeOf(GetType(PRINTER_NOTIFY_INFO))
        Dim data As PRINTER_NOTIFY_INFO_DATA() = New PRINTER_NOTIFY_INFO_DATA(info.Count) {}
        Dim i As UInteger = 0
        While i < info.Count
            data(i) = CType(Marshal.PtrToStructure(CType(pData, IntPtr), GetType(PRINTER_NOTIFY_INFO_DATA)), PRINTER_NOTIFY_INFO_DATA)
            pData += Marshal.SizeOf(GetType(PRINTER_NOTIFY_INFO_DATA))
            i += 1
        End While
#End Region

#Region "iterate through all elements in the data array"
        i = 0
        While i < data.Count()
            Dim jStatus As JOBSTATUS = CType([Enum].Parse(GetType(JOBSTATUS), data(i).NotifyData.Data.cbBuf.ToString()), JOBSTATUS)
            Dim intJobID As Integer = CType(data(i).Id, Integer)
            Dim strJobName As String = ""
            Dim pji As PrintSystemJobInfo = Nothing
            If (data(i).Field = CType(PRINTERJOBNOTIFICATIONTYPES.JOB_NOTIFY_FIELD_STATUS, UShort)) AndAlso (data(i).Type = CType(PRINTERNOTIFICATIONTYPES.JOB_NOTIFY_TYPE, UShort)) Then
                'Dim jStatus As JOBSTATUS = CType([Enum].Parse(GetType(JOBSTATUS), data(i).NotifyData.Data.cbBuf.ToString()), JOBSTATUS)
                'Dim intJobID As Integer = CType(data(i).Id, Integer)
                'Dim strJobName As String = ""
                'Dim pji As PrintSystemJobInfo = Nothing
                Try
                    _spooler = New PrintQueue(New PrintServer(), _spoolerName)
                    pji = _spooler.GetJob(intJobID)
                    If Not objJobDict.ContainsKey(intJobID) Then
                        objJobDict(intJobID) = pji.Name
                    End If
                    strJobName = pji.Name
                Catch
                    pji = Nothing
                    objJobDict.TryGetValue(intJobID, strJobName)
                    If strJobName Is Nothing Then
                        strJobName = ""
                    End If
                End Try

                'Let us raise the event
                'RaiseEvent OnJobStatusChange(Me, New PrintJobChangeEventArgs(intJobID, strJobName, jStatus, pji))
            End If
            RaiseEvent OnJobStatusChange(Me, New PrintJobChangeEventArgs(intJobID, strJobName, jStatus, pji))
            i += 1
        End While
#End Region

#Region "reset the Event and wait for the next event"
        _mrEvent.Reset()
        ' _waitHandle = ThreadPool.RegisterWaitForSingleObject(_mrEvent, New WaitOrTimerCallback(AddressOf PrinterNotifyWaitCallback), _mrEvent, -1, true)
#End Region

    End Sub
    '#End Region


End Class

